! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

program fd4_demo_heat

  use fd4_mod
  implicit none
  include 'mpif.h'
  
  ! Setup parameters
  !
  real, parameter :: radius = 0.5        ! rel. radius of initial heat pertubation
  integer, parameter :: grid(3) = 32     ! number of grid cells
  real(r_k), parameter :: ds(3) = 1.0    ! grid cell size
  real(r_k), parameter :: dt = 0.1       ! time step size
  integer, parameter :: nsteps = 1200    ! number of time steps to compute
  integer, parameter :: outputsteps = 12 ! number of tite steps with output (console, Vis5D)

  ! Variables
  !
  ! FD4 variable table
  type(fd4_vartab) :: vartab(1)
  integer, parameter :: THETA = 1
  ! FD4 domain
  type(fd4_domain), target :: domain
  integer :: dsize(3,2), bnum(3), nghosts(3)
  logical :: periodic(3)
  ! FD4 iterator
  type(fd4_iter) :: iter
  ! FD4 ghost communication
  type(fd4_ghostcomm) :: ghostcomm(2)
  ! misc
  integer :: rank, nproc, err, bext(0:3)
  real(r_k), allocatable :: buffer(:,:,:,:)
  real(r_k) :: dtheta
  integer :: offset(3), x, y, z, output_modulo
  real :: global_pos(3), cr
  integer :: now, new, step
  integer(i8k) :: t0, t1


  ! MPI Initialization
  !
  call MPI_Init(err)
  call MPI_Comm_rank(MPI_COMM_WORLD,rank ,err) 
  call MPI_Comm_size(MPI_COMM_WORLD,nproc,err)

  ! Create the FD4 variable table
  !  (only one cell-centered variable 'theta' with 2 time steps)
  !
  !                             name, bins, steps,  unused,   vnull,      vthres, cell-cent.
  vartab(THETA) = fd4_vartab('theta',    1,     2, .false., 0.0_r_k, FD4_NOTHRES, FD4_CELLC )

  ! Create the FD4 domain
  !
  dsize(1:3,1) = (/1, 1, 1/)    ! grid start indices
  dsize(1:3,2) = grid(1:3)      ! grid end   indices
  bnum(1:3) = grid(1:3) / 2     ! number of blocks in each dimension
  nghosts(1:3) = (/1, 1, 1/)    ! number of ghost cells in each dimension
  periodic(1:3) = .true.        ! periodic boundaries
  call fd4_domain_create(domain, bnum, dsize, vartab, nghosts, periodic, MPI_COMM_WORLD, err)
  if(err/=0) then
    write(*,*) rank, ': fd4_domain_create failed'
    call MPI_Abort(MPI_COMM_WORLD, 1, err)
  end if
  
  ! Allocate the buffer array for a single block inclusive ghost cells
  !
  ! get the max block extent (bext) incl. ghost cells
  call fd4_domain_max_bext(domain, bext(1:3), .true.)
  bext(0) = 1 ! "bin" dimension (not used here)
  ! allocate the buffer array
  allocate( buffer(bext(0),0:bext(1)-1,0:bext(2)-1,0:bext(3)-1) )
  buffer = 0.0_r_k
  
  ! Allocate the blocks of the domain
  !
  ! allocated all blocks balanced over all processes (Hilbert SFC partitioning)
  call fd4_util_allocate_all_blocks(domain, err)
  
  ! Initialize time step indicators
  !
  now = 1
  new = 2
    
  ! Initialize the domain variables
  ! 
  ! iterate over all local blocks
  call fd4_iter_init(domain, iter)
  do while(associated(iter%cur))
    ! offset from domain indexes to block-local indexes
    call fd4_iter_offset(iter, offset)
    ! loop over block's grid cells
    do z=1,iter%cur%ext(3)
      do y=1,iter%cur%ext(2)
        do x=1,iter%cur%ext(1)
          ! get global coordinates of this grid cell and scale to [0,1]
          global_pos(1) = REAL( offset(1) + x - 1 ) / REAL(dsize(1,2) - 1)
          global_pos(2) = REAL( offset(2) + y - 1 ) / REAL(dsize(2,2) - 1)
          global_pos(3) = REAL( offset(3) + z - 1 ) / REAL(dsize(3,2) - 1)
          ! distance from domain center to current grid cell
          cr = sqrt( (global_pos(1)-0.5)**2 + (global_pos(2)-0.5)**2 + (global_pos(3)-0.5)**2 )
          if( cr < radius ) iter%cur%fields(THETA,now)%l(1,x,y,z) = 2.0_r_k * cos(3.14159*cr/(2*radius))
        end do
      end do
    end do    
    call fd4_iter_next(iter)
  end do
  
  ! Create ghost communicator for variable THETA
  !  (one for each time level)
  !
  call fd4_ghostcomm_create(ghostcomm(1), domain, 1, (/THETA/), (/1/), err)
  call fd4_ghostcomm_create(ghostcomm(2), domain, 1, (/THETA/), (/2/), err)

  
  ! Initialize Output
  !
  output_modulo = 2 * nsteps
  if(outputsteps>0) then
    output_modulo = nsteps / outputsteps
#ifdef VIS5D
    call fd4_vis5d_open(domain,'out.v5d',outputsteps+1,1,(/THETA/),(/1/),err)  
    call fd4_vis5d_write(err,(/THETA/),(/now/))
#endif
  end if
  
  ! Get time before time stepping
  !  (gettime is part of FD4's utilities)
  !
  call gettime(t0)
  
  ! Time stepping loop
  !
  do step = 1,nsteps
    ! exchange ghost cells for time level 'now'
    call fd4_ghostcomm_exch(ghostcomm(now), err)
    ! iterate over all local blocks
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))
      ! get theta with ghost cells from current block
      call fd4_iter_get_ghost(iter, THETA, now, bext, buffer)
      ! loop over block's grid cells
      do z=1,iter%cur%ext(3)
        do y=1,iter%cur%ext(2)
          do x=1,iter%cur%ext(1)
            !                           ( d2T   d2T   d2T )
            !  theta_new =  theta_now + ( --- + --- + --- ) * dt
            !                           ( dx2   dy2   dz2 )
            ! calculate delta
            dtheta = ( buffer(1,x-1,y,z) + buffer(1,x+1,y,z) - 2*buffer(1,x,y,z) ) / (ds(1) * ds(1)) &
                   + ( buffer(1,x,y-1,z) + buffer(1,x,y+1,z) - 2*buffer(1,x,y,z) ) / (ds(2) * ds(2)) &
                   + ( buffer(1,x,y,z-1) + buffer(1,x,y,z+1) - 2*buffer(1,x,y,z) ) / (ds(3) * ds(3))
            ! set updated theta value
            iter%cur%fields(THETA,new)%l(1,x,y,z) = buffer(1,x,y,z) + dtheta * dt
          end do
        end do
      end do    
      call fd4_iter_next(iter)
    end do
    ! end of block loop
    ! output
    if(mod(step,output_modulo)==0) then
      if(rank==0) write(*,'(A,I5)') 'step: ',step
#ifdef VIS5D
      ! write Vis5D output
      if(outputsteps>0) call fd4_vis5d_write(err,(/THETA/),(/new/))
#endif
    end if
    ! swap time step indicators
    now = 3 - now
    new = 3 - new  
  end do
  
  ! Get time after time stepping
  !
  call gettime(t1)  
  if(rank==0) write(*,'(A,F12.4,A)') 'time: ', REAL(t1-t0)/1.e6, ' s'  
  
  ! Close Vis5D
  !
#ifdef VIS5D
  if(outputsteps>0) call fd4_vis5d_close(err)
#endif

  ! Delete the ghost communicator and the domain
  !
  call fd4_ghostcomm_delete(ghostcomm(1))
  call fd4_ghostcomm_delete(ghostcomm(2))
  call fd4_domain_delete(domain)

  call MPI_Finalize(err)

end program fd4_demo_heat
